from autonr.autonr import AutoNR
from autonr.nrkmeans import _get_precision
from sklearn.metrics import normalized_mutual_info_score as nmi
import numpy as np
from sklearn.preprocessing import scale
from sklearn.decomposition import PCA
import time
from scipy.spatial.distance import cdist

"""
Helpers
"""


def _check_number_of_points(labels_true, labels_pred):
    if labels_pred.shape[0] != labels_true.shape[0]:
        raise Exception(
            "Number of objects of the prediction and ground truth are not equal.\nNumber of prediction objects: " + str(
                labels_pred.shape[0]) + "\nNumber of ground truth objects: " + str(labels_true.shape[0]))


def remove_noise_spaces_from_labels(labels):
    no_noise_spaces = [True] * labels.shape[1]
    for c in range(labels.shape[1]):
        unique_labels = np.unique(labels[:, c])
        # Consider outliers
        len_unique_labels = len(unique_labels) if np.all(unique_labels >= 0) else len(unique_labels) - 1
        if len_unique_labels == 1:
            no_noise_spaces[c] = False
    labels_new = labels[:, no_noise_spaces]
    return labels_new


class MultipleLabelingsConfusionMatrix():

    def __init__(self, labels_true, labels_pred, metric=nmi, remove_noise_spaces=True, metric_params={}):
        _check_number_of_points(labels_true, labels_pred)
        if labels_true.ndim == 1:
            labels_true = labels_true.reshape((-1, 1))
        if labels_pred.ndim == 1:
            labels_pred = labels_pred.reshape((-1, 1))
        if remove_noise_spaces:
            labels_true = remove_noise_spaces_from_labels(labels_true)
            labels_pred = remove_noise_spaces_from_labels(labels_pred)
        if labels_true.shape[1] == 0 or labels_pred.shape[1] == 0:
            raise Exception("labels_true or labels_pred matrix contains zero columns.")
        assert type(metric_params) is dict, "metric_params must be a dict"
        assert callable(metric), "metric must be a method"
        confusion_matrix = np.zeros((labels_true.shape[1], labels_pred.shape[1]))
        for i in range(labels_true.shape[1]):
            for j in range(labels_pred.shape[1]):
                confusion_matrix[i, j] = metric(labels_true[:, i], labels_pred[:, j], **metric_params)
        self.confusion_matrix = confusion_matrix

def _precision_score(n_tp, n_fp):
    return 0 if (n_tp + n_fp) == 0 else n_tp / (n_tp + n_fp)


def _recall_score(n_tp, n_fn):
    return 0 if (n_tp + n_fn) == 0 else n_tp / (n_tp + n_fn)


def f1_score(labels_true, labels_pred):
    _check_number_of_points(labels_true, labels_pred)
    if labels_true.ndim != 1 or labels_pred.ndim != 1:
        raise Exception("labels_true and labels_pred labels should just contain a single column.")
    n_tp = 0
    n_fp = 0
    n_fn = 0
    n_tn = 0
    for i in range(labels_pred.shape[0] - 1):
        n_tp += np.sum((labels_pred[i] == labels_pred[i + 1:]) & (labels_true[i] == labels_true[i + 1:]))
        n_fp += np.sum((labels_pred[i] == labels_pred[i + 1:]) & (labels_true[i] != labels_true[i + 1:]))
        n_fn += np.sum((labels_pred[i] != labels_pred[i + 1:]) & (labels_true[i] == labels_true[i + 1:]))
        n_tn += np.sum((labels_pred[i] != labels_pred[i + 1:]) & (labels_true[i] != labels_true[i + 1:]))
    precision = _precision_score(n_tp, n_fp)
    recall = _recall_score(n_tp, n_fn)
    return 0 if (precision == 0 and recall == 0) else 2 * precision * recall / (precision + recall)


"""
Experiments
"""


def evaluate(labels, prediction, runtime):
    print("- Runtime\n", runtime)
    mc_nmi = MultipleLabelingsConfusionMatrix(labels, prediction, nmi)
    mc_nmi_max = np.max(mc_nmi.confusion_matrix, axis=1)
    print("- NMI\n", mc_nmi_max)
    mc_f1 = MultipleLabelingsConfusionMatrix(labels, prediction, f1_score)
    mc_f1_max = np.max(mc_f1.confusion_matrix, axis=1)
    print("- F1\n", mc_f1_max)
    return mc_nmi_max, mc_f1_max


def execute_test(data, labels, repetitions, get_max_and_precision_first=True):
    if get_max_and_precision_first:
        max_distance = np.max(cdist(data, data))
        precision = _get_precision(data)
    else:
        max_distance, precision = None, None
    for outliers in [False, True]:
        labels_shape = 1 if labels.ndim == 1 else labels.shape[1]
        nmi_results = np.zeros((repetitions, labels_shape))
        f1_results = np.zeros((repetitions, labels_shape))
        for i in range(repetitions):
            print("===> {0} outliers: {1}". format("With" if outliers else "Without", i))
            begin = time.time()
            result = AutoNR(outliers=outliers, debug=False, nrkmeans_repetitions=15, max_distance=max_distance,
                            precision=precision)
            result.fit(data)
            end = time.time()
            mc_nmi, mc_f1 = evaluate(labels, result.nrkmeans.labels_, end - begin)
            nmi_results[i] = mc_nmi
            f1_results[i] = mc_f1
        print("NMI  -  Mean: {0} / Std: {1}".format(np.mean(nmi_results, axis=0), np.std(nmi_results, axis=0)))
        print("F1  -  Mean: {0} / Std: {1}".format(np.mean(f1_results, axis=0), np.std(f1_results, axis=0)))


def runtime_experiments(repetitions=10):
    print("==== N5000 ====")
    dataset = np.genfromtxt("autonr/data/N5000.csv", delimiter=",")
    data = dataset[:, 3:]
    labels = dataset[:, :3]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions, False)

    print("==== N10000 ====")
    dataset = np.genfromtxt("autonr/data/N10000.csv", delimiter=",")
    data = dataset[:, 3:]
    labels = dataset[:, :3]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions, False)

    print("==== N15000 ====")
    dataset = np.genfromtxt("autonr/data/N15000.csv", delimiter=",")
    data = dataset[:, 3:]
    labels = dataset[:, :3]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions, False)

    print("==== N20000 ====")
    dataset = np.genfromtxt("autonr/data/N20000.csv", delimiter=",")
    data = dataset[:, 3:]
    labels = dataset[:, :3]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions, False)

    print("==== N25000 ====")
    dataset = np.genfromtxt("autonr/data/N25000.csv", delimiter=",")
    data = dataset[:, 3:]
    labels = dataset[:, :3]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions, False)

    print("==== d16 ====")
    dataset = np.genfromtxt("autonr/data/d16.csv", delimiter=",")
    data = dataset[:, 3:]
    labels = dataset[:, :3]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions, False)

    print("==== d26 ====")
    dataset = np.genfromtxt("autonr/data/d26.csv", delimiter=",")
    data = dataset[:, 3:]
    labels = dataset[:, :3]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions, False)

    print("==== d36 ====")
    dataset = np.genfromtxt("autonr/data/d36.csv", delimiter=",")
    data = dataset[:, 3:]
    labels = dataset[:, :3]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions, False)

    print("==== d46 ====")
    dataset = np.genfromtxt("autonr/data/d46.csv", delimiter=",")
    data = dataset[:, 3:]
    labels = dataset[:, :3]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions, False)

    print("==== J5 ====")
    dataset = np.genfromtxt("autonr/data/J5.csv", delimiter=",")
    data = dataset[:, 5:]
    labels = dataset[:, :5]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions, False)

    print("==== J7 ====")
    dataset = np.genfromtxt("autonr/data/J7.csv", delimiter=",")
    data = dataset[:, 7:]
    labels = dataset[:, :7]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions, False)

    print("==== J9 ====")
    dataset = np.genfromtxt("autonr/data/J9.csv", delimiter=",")
    data = dataset[:, 9:]
    labels = dataset[:, :9]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions, False)

    print("==== J11 ====")
    dataset = np.genfromtxt("autonr/data/J11.csv", delimiter=",")
    data = dataset[:, 11:]
    labels = dataset[:, :11]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions, False)


def qualitative_experiments(repetitions=10):
    np.random.seed(11)

    print("==== SYN3 ====")
    dataset = np.genfromtxt("autonr/data/syn3.csv", delimiter=",")
    data = dataset[:, 3:]
    labels = dataset[:, :3]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions)

    print("==== SYN3o ====")
    dataset = np.genfromtxt("autonr/data/syn3o.csv", delimiter=",")
    data = dataset[:, 3:]
    labels = dataset[:, :3]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions)

    print("==== fruits ====")
    dataset = np.genfromtxt("autonr/data/fruit.data", delimiter=",")
    data = dataset[:, 2:]
    labels = dataset[:, :2]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    execute_test(data, labels, repetitions)

    print("==== ALOI ====")
    dataset = np.genfromtxt("autonr/data/aloi.data", delimiter=";")
    data = dataset[:, 2:]
    labels = dataset[:, :2]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    pca = PCA(0.9)
    data = pca.fit_transform(data)
    execute_test(data, labels, repetitions)

    print("==== DSF ====")
    dataset = np.genfromtxt("autonr/data/dancing_stickfigures.data", delimiter=";")
    data = dataset[:, 3:]
    labels = dataset[:, :2]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    pca = PCA(0.9)
    data = pca.fit_transform(data)
    execute_test(data, labels, repetitions)

    print("==== CMUface ====")
    dataset = np.genfromtxt("autonr/data/cmuface.csv", delimiter=",")
    data = dataset[:, 2:]
    labels = dataset[:, :2]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    pca = PCA(0.9)
    data = pca.fit_transform(data)
    execute_test(data, labels, repetitions)

    print("==== WebKB ====")
    dataset = np.genfromtxt("autonr/data/webkb.csv", delimiter=",")
    data = dataset[:, 2:]
    labels = dataset[:, :2]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    pca = PCA(0.9)
    data = pca.fit_transform(data)
    execute_test(data, labels, repetitions)

    print("==== NRLetters ====")
    dataset = np.genfromtxt("autonr/data/nrLetters.csv", delimiter=",")
    data = dataset[:, :-3]
    labels = dataset[:, -3:]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    pca = PCA(0.9)
    data = pca.fit_transform(data)
    execute_test(data, labels, repetitions)

    print("==== Wine ====")
    dataset = np.genfromtxt("autonr/data/wine.csv", delimiter=",")
    data = dataset[:, 1:]
    labels = dataset[:, :1]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    data = scale(data, axis=0)
    execute_test(data, labels, repetitions)

    print("==== Optdigits ====")
    dataset = np.genfromtxt("autonr/data/optdigits.csv", delimiter=",")
    data = dataset[:, 1:]
    labels = dataset[:, :1]
    print("Data shape: {0} / Labels shape: {1}".format(data.shape, labels.shape))
    pca = PCA(0.9)
    data = pca.fit_transform(data)
    execute_test(data, labels, repetitions)
